<?php
/**
 * Hermes storage implementation for PHP's PEAR database abstraction layer.
 *
 * $Horde: hermes/lib/Driver/sql.php,v 1.27 2004/12/10 18:07:46 jan Exp $
 *
 * Required values for $params:<pre>
 *      'phptype'       The database type (e.g. 'pgsql', 'mysql, etc.).
 *      'hostspec'      The hostname of the database server.
 *      'protocol'      The communication protocol ('tcp', 'unix', etc.).
 *      'username'      The username with which to connect to the database.
 *      'password'      The password associated with 'username'.
 *      'database'      The name of the database.
 *
 * Required by some database implementations:
 *      'options'       Additional options to pass to the database.
 *      'tty'           The TTY on which to connect to the database.
 *      'port'          The port on which to connect to the database.
 *
 * The table structure can be created by the
 * scripts/drivers/hermes.sql script.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org
 * @version $Revision: 1.27 $
 * @since   Hermes 0.0.1
 * @package Hermes
 */
class Hermes_Driver_sql extends Hermes_Driver {

    /** Hash containing connection parameters. */
    var $_params = array();

    /** Handle for the current database connection.
        @var object DB $db */
    var $_db;

    /** Boolean indicating whether or not we're connected to the SQL server. */
    var $_connected = false;

    /**
     * Constructs a new SQL storage object.
     *
     * @param string $user      The user who owns these billing.
     * @param array  $params    A hash containing connection parameters.
     */
    function Hermes_Driver_sql($params = array())
    {
        $this->_params = $params;
    }

    /**
     * Save a row of billing information.
     *
     * @access public
     *
     * @param string  $employee  The Horde ID of the person who worked the hours.
     * @param array   $entries   The billing information to enter. Each array row
     *                           must contain the following entries:
     *            'date'         The day the hours were worked (ISO format)
     *            'client'       The id of the client the work was done for.
     *            'type'         The type of work done.
     *            'hours'        The number of hours worked
     *            'rate'         The hourly rate the work was done at.
     *            'billable'     (optional) Whether or not the work is billable hours.
     *            'description'  A short description of the work.
     *
     * @return mixed  True on success, PEAR_Error on failure.
     */
    function enterTime($employee, $info)
    {
        require_once 'Date.php';

        /* Make sure we have a valid database connection. */
        $this->_connect();

        $dt = &new Date($info['date']);
        $timeslice_id = $this->_db->nextId('hermes_timeslices');
        $sql = sprintf('INSERT INTO hermes_timeslices (timeslice_id, clientjob_id, employee_id, jobtype_id,' .
                       ' timeslice_hours, timeslice_isbillable, timeslice_date, timeslice_description, timeslice_note)' .
                       ' VALUES (%d, %s, %s, %s, %s, %s, %d, %s, %s)',
                       (int)$timeslice_id,
                       $this->_db->quote($info['client']),
                       $this->_db->quote($employee),
                       $this->_db->quote($info['type']),
                       $this->_db->quote($info['hours']),
                       isset($info['billable']) ? (int)$info['billable'] : 0,
                       (int)$dt->getTime(DATE_FORMAT_UNIXTIME) + 1,
                       $this->_db->quote($info['description']),
                       $this->_db->quote($info['note']));

        Horde::logMessage($sql, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $result = $this->_db->query($sql);
        if (DB::isError($result)) {
            return $result;
        }

        return true;
    }

    /**
     * Update a set of billing information.
     *
     * @access public
     *
     * @param array   $entries   The billing information to enter. Each array row
     *                           must contain the following entries:
     *            'id'           The id of this time entry.
     *            'date'         The day the hours were worked (ISO format)
     *            'client'       The id of the client the work was done for.
     *            'type'         The type of work done.
     *            'hours'        The number of hours worked
     *            'rate'         The hourly rate the work was done at.
     *            'billable'     Whether or not the work is billable hours.
     *            'description'  A short description of the work.
     *
     *                           If any rows contain a 'delete' entry, those rows will
     *                           be deleted instead of updated.
     *
     * @return mixed  True on success, PEAR_Error on failure.
     */
    function updateTime($entries)
    {
        require_once 'Date.php';

        /* Make sure we have a valid database connection. */
        $this->_connect();

        foreach ($entries as $info) {
            if (!Hermes::canEditTimeslice($info['id'])) {
                return PEAR::raiseError(_("Access denied; user cannot modify this timeslice."));
            }
            if (!empty($info['delete'])) {
                $sql = sprintf('DELETE FROM hermes_timeslices' .
                               ' WHERE timeslice_id = %d',
                               (int)$info['id']);
            } else {
                if (isset($info['employee'])) {
                    $employee_cl = sprintf(' employee_id = %s,',
                                           $this->_db->quote($info['employee']));
                } else {
                    $employee_cl = '';
                }
                $dt = &new Date($info['date']);
                $sql = sprintf('UPDATE hermes_timeslices SET' .
                               $employee_cl .
                               ' clientjob_id = %s, jobtype_id = %s,' .
                               ' timeslice_hours = %s, timeslice_isbillable = %d,' .
                               ' timeslice_date = %d, timeslice_description = %s, timeslice_note = %s' .
                               ' WHERE timeslice_id = %d',
                               $this->_db->quote($info['client']),
                               $this->_db->quote($info['type']),
                               $this->_db->quote($info['hours']),
                               isset($info['billable']) ? (int)$info['billable'] : 0,
                               (int)$dt->getTime(DATE_FORMAT_UNIXTIME) + 1,
                               $this->_db->quote($info['description']),
                               $this->_db->quote($info['note']),
                               (int)$info['id']);
            }

            Horde::logMessage($sql, __FILE__, __LINE__, PEAR_LOG_DEBUG);
            $result = $this->_db->query($sql);
            if (DB::isError($result)) {
                return $result;
            }
        }

        return true;
    }

    function getHours($filters = array())
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        $sql = sprintf('SELECT b.timeslice_id as id,' .
                       ' b.clientjob_id as client,' .
                       ' b.employee_id as employee,' .
                       ' b.jobtype_id as type,' .
                       ' j.jobtype_name as "_type_name",' .
                       ' b.timeslice_hours as hours,' .
                       ' b.timeslice_rate as rate,' .
                       ' b.timeslice_isbillable as billable,' .
                       ' b.timeslice_date as "date",' .
                       ' b.timeslice_description as description,' .
                       ' b.timeslice_note as note,' .
                       ' b.timeslice_submitted as submitted' .
                       ' FROM hermes_timeslices b INNER JOIN hermes_jobtypes j ON b.jobtype_id = j.jobtype_id');
        if (count($filters) > 0) {
            $sql .= ' WHERE';
            $glue = '';
            foreach ($filters as $field => $filter) {
                switch ($field) {
                case 'client':
                    $sql .= $glue . $this->_equalClause('b.clientjob_id',
                                                        $filter);
                    $glue = ' AND';
                    break;

                case 'jobtype':
                    $sql .= $glue . $this->_equalClause('b.jobtype_id',
                                                        $filter);
                    $glue = ' AND';
                    break;

                case 'submitted':
                    $sql .= $glue . ' timeslice_submitted = ' . (int)$filter;
                    $glue = ' AND';
                    break;

                case 'exported':
                    $sql .= $glue . ' timeslice_exported = ' . (int)$filter;
                    $glue = ' AND';
                    break;

                case 'start':
                    $sql .= $glue . ' timeslice_date >= ' . (int)$filter;
                    $glue = ' AND';
                    break;

                case 'end':
                    $sql .= $glue . ' timeslice_date <= ' . (int)$filter;
                    $glue = ' AND';
                    break;

                case 'employee':
                    $sql .= $glue . $this->_equalClause('employee_id',
                                                        $filter);
                    $glue = ' AND';
                    break;

                case 'id':
                    $sql .= $glue . $this->_equalClause('timeslice_id',
                                                        $filter, false);
                    $glue .= ' AND';
                    break;
                }
            }
        }

        $sql .= ' ORDER BY timeslice_date, clientjob_id';

        Horde::logMessage($sql, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        $hours = $this->_db->getAll($sql, DB_FETCHMODE_ASSOC);

        return $hours;
    }

    /**
     * @access private
     */
    function _equalClause($lhs, $rhs, $quote = true)
    {
        require_once 'Horde/SQL.php';

        if (!is_array($rhs)) {
            if ($quote) {
                return sprintf(' %s = %s', $lhs, $this->_db->quote($rhs));
            }
            return sprintf(' %s = %s', $lhs, $rhs);
        }

        if (count($rhs) == 0) {
            return ' FALSE';
        }

        $glue = '';
        $ret = sprintf(' %s IN ( ', $lhs);
        foreach ($rhs as $value) {
            $ret .= $glue . $this->_db->quote($value);
            $glue = ', ';
        }
        return $ret . ' )';
    }

    function markAs($field, $hours)
    {
        if (!count($hours)) {
            return false;
        }

        $this->_connect();

        switch ($field) {
        case 'submitted':
            $h_field = 'timeslice_submitted';
            break;

        case 'exported':
            $h_field = 'timeslice_exported';
            break;

        default:
            return false;
        }

        $ids = array();
        foreach ($hours as $entry) {
            $ids[] = (int)$entry['id'];
        }

        $sql = 'UPDATE hermes_timeslices SET ' . $h_field . ' = 1 WHERE timeslice_id IN (' . implode(',', $ids) . ')';
        Horde::logMessage($sql, __FILE__, __LINE__, PEAR_LOG_DEBUG);
        return $this->_db->query($sql);
    }

    function addJobType($name)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        // Get a new unique id.
        $new_id = $this->_db->nextId('hermes_jobtypes');
        if (DB::isError($new_id)) {
            return $new_id;
        }

        $result = $this->_db->query('INSERT INTO hermes_jobtypes (jobtype_id, jobtype_name) VALUES (' .
                                    $this->_db->quote($new_id) . ', ' . $this->_db->quote($name) . ')');
        if (DB::isError($result)) {
            return $result;
        }

        return $new_id;
    }

    function getJobType($jobTypeID)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        if (empty($jobTypeID)) {
            return false;
        }
        $jobType = $this->_db->getAssoc('SELECT jobtype_id, jobtype_name FROM hermes_jobtypes WHERE jobtype_id = ' . $this->_db->quote($jobTypeID));
        if (DB::isError($jobType)) {
            return $jobType;
        }
        return array('id' => $jobTypeID,
                     'name' => $jobType[$jobTypeID]);
    }

    function getJobTypes()
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        return $this->_db->getAssoc('SELECT jobtype_id, jobtype_name FROM hermes_jobtypes ORDER BY jobtype_name');
    }

    function updateJobType($jobTypeID, $name)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        $result = $this->_db->query('UPDATE hermes_jobtypes SET' .
                                    ' jobtype_name = ' . $this->_db->quote($name) .
                                    ' WHERE jobtype_id = ' . $this->_db->quote($jobTypeID));

        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        return true;
    }

    function deleteJobType($jobTypeID)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        return $this->_db->query('DELETE FROM hermes_jobtypes WHERE jobtype_id = ' . $this->_db->quote($jobTypeID));
    }

    function getClientSettings($clientID)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        $clients = Hermes::listClients();
        if (empty($clientID) || !isset($clients[$clientID])) {
            return PEAR::raiseError('Does not exist');
        }
        $settings = $this->_db->getAssoc('SELECT clientjob_id, clientjob_enterdescription, clientjob_exportid FROM hermes_clientjobs WHERE clientjob_id = ' . $this->_db->quote($clientID));
        if (is_a($settings, 'PEAR_Error')) {
            return $settings;
        }

        if (isset($clientJob[$clientID])) {
            $settings = array('id' => $clientID,
                              'enterdescription' => $clientJob[$clientID][0],
                              'exportid' => $clientJob[$clientID][1]);
        } else {
            $settings = array('id' => $clientID,
                              'enterdescription' => 1,
                              'exportid' => null);
        }

        $settings['name'] = $clients[$clientID];
        return $settings;
    }

    function updateClientSettings($clientID, $enterdescription = 1, $exportid = null)
    {
        /* Make sure we have a valid database connection. */
        $this->_connect();

        if (empty($exportid)) {
            $exportid = null;
        }

        $exists = 'SELECT clientjob_id FROM hermes_clientjobs WHERE clientjob_id = ' . $this->_db->quote($clientID);
        if ($this->_db->getOne($exists) !== $clientID) {
            $result = $this->_db->query('INSERT INTO hermes_clientjobs (clientjob_id, clientjob_enterdescription, clientjob_exportid) VALUES (' .
                                        $this->_db->quote($clientID) . ', ' . (int)$enterdescription . ', ' . $this->_db->quote($exportid) . ')');
        } else {
            $result = $this->_db->query('UPDATE hermes_clientjobs SET' .
                                        ' clientjob_exportid = ' . $this->_db->quote($exportid) . ', ' .
                                        ' clientjob_enterdescription = ' . (int)$enterdescription .
                                        ' WHERE clientjob_id = ' . $this->_db->quote($clientID));
        }

        return PEAR::isError($result) ? $result : true;
    }

    function purge()
    {
        global $conf;

        /* Make sure we have a valid database connection. */
        $this->_connect();

        $query = sprintf('DELETE FROM hermes_timeslices WHERE timeslice_exported = 1 AND timeslice_date < %s',
                         mktime(0, 0, 0, date('n'), date('j') - $conf['time']['days_to_keep']));
        return $this->_db->query($query);
    }

    /**
     * Attempts to open a persistent connection to the SQL server.
     *
     * @return boolean    True on success; exits (Horde::fatal()) on error.
     */
    function _connect()
    {
        if (!$this->_connected) {
            Horde::assertDriverConfig($this->_params, 'storage',
                array('phptype', 'hostspec', 'username', 'database'));

            /* Connect to the SQL server using the supplied parameters. */
            include_once 'DB.php';
            $this->_db = &DB::connect($this->_params,
                                      array('persistent' => !empty($this->_params['persistent'])));
            if (is_a($this->_db, 'PEAR_Error')) {
                Horde::fatal($this->_db, __FILE__, __LINE__);
            }

            $this->_db->setOption('portability', DB_PORTABILITY_LOWERCASE | DB_PORTABILITY_ERRORS);

            $this->_connected = true;
        }

        return true;
    }

    /**
     * Disconnect from the SQL server and clean up the connection.
     *
     * @return boolean     true on success, false on failure.
     */
    function _disconnect()
    {
        if ($this->_connected) {
            $this->_connected = false;
            return $this->_db->disconnect();
        }

        return true;
    }

}
